@@ -28,6 +28,7 @@ import tempfile
 import urllib2
 import mimetypes        # better use the magic packet, but that's not a standard lib
 import gtkgui_helpers
+import threading
 from Queue import Queue
 try:
     from PIL import Image
@@ -83,16 +84,21 @@ class HttpuploadPlugin(GajimPlugin):
             self.available_text = DEP_MSG
         self.config_dialog = None  # HttpuploadPluginConfigDialog(self)
         self.controls = []
-        self.events_handlers = {}
-        self.events_handlers['agent-info-received'] = (ged.PRECORE,
-                self.handle_agent_info_received)
-        self.events_handlers['raw-iq-received'] = (ged.PRECORE,
-                self.handle_iq_received)
+        self.events_handlers = {
+            'agent-info-received': (
+                ged.PRECORE, self.handle_agent_info_received),
+            'stanza-message-outgoing': (
+                99, self.handle_outgoing_stanza),
+            'gc-stanza-message-outgoing': (
+                99, self.handle_outgoing_stanza),
+            'raw-iq-received': (
+                ged.PRECORE, self.handle_iq_received)}
         self.gui_extension_points = {
             'chat_control_base': (self.connect_with_chat_control,
                 self.disconnect_from_chat_control),
             'chat_control_base_update_toolbar': (self.update_button_state,
                 None)}
+        self.messages = []
         self.first_run = True
 
     def handle_iq_received(self, event):
@@ -116,6 +122,13 @@ class HttpuploadPlugin(GajimPlugin):
             for base in self.controls:
                 self.update_button_state(base.chat_control)
 
+    def handle_outgoing_stanza(self, event):
+        message = event.msg_iq.getTagData('body')
+        if message and message in self.messages:
+            self.messages.remove(message)
+            oob = event.msg_iq.addChild('x', namespace=nbxmpp.NS_X_OOB)
+            oob.addChild('url').setData(message)
+
     @log_calls('HttpuploadPlugin')
     def connect_with_chat_control(self, control):
         self.chat_control = control
@@ -276,19 +289,14 @@ class Base(object):
             del self.chat_control.handlers[self.keypress_id]
 
     def encryption_activated(self):
-        if not encryption_available:
-            return False
         jid = self.chat_control.contact.jid
         account = self.chat_control.account
         for plugin in gajim.plugin_manager.active_plugins:
             if type(plugin).__name__ == 'OmemoPlugin':
-                omemo = plugin
-                break
-        if omemo:
-            state = omemo.get_omemo_state(account)
-            log.info('Encryption is: ' +
-                      str(state.encryption.is_active(jid)))
-            return state.encryption.is_active(jid)
+                state = plugin.get_omemo_state(account)
+                log.info('Encryption is: ' +
+                          str(state.encryption.is_active(jid)))
+                return state.encryption.is_active(jid)
         log.info('Encryption is: False / OMEMO not found')
         return False
 
@@ -301,6 +309,13 @@ class Base(object):
             log.debug(e)
             self.encrypted_upload = False
 
+        if self.encrypted_upload and not encryption_available:
+            ErrorDialog(
+                _('Error'),
+                'Please install python-cryptography for encrypted uploads',
+                transient_for=self.chat_control.parent_win.window)
+            return
+
         if not path_to_file:
             path_to_file = self.dlg.get_filename()
             if not path_to_file:
@@ -334,21 +349,28 @@ class Base(object):
             mime_type = 'application/octet-stream'  # fallback mime type
         log.info("Detected MIME Type of file: " + str(mime_type))
         progress_messages = Queue(8)
-        progress_window = ProgressWindow(_('HTTP Upload'), _('Requesting HTTP Upload Slot...'), progress_messages, self.plugin)
+        event = threading.Event()
+        progress_window = ProgressWindow(_('Requesting HTTP Upload Slot...'), progress_messages, self, event)
+
         def upload_file(stanza):
-            slot = stanza.getTag("slot")
-            if not slot:
+            if stanza.getType() == 'error':
+                ErrorDialog(_('Could not request upload slot'),
+                            stanza.getErrorMsg(),
+                            transient_for=self.chat_control.parent_win.window)
+                log.error(stanza)
                 progress_window.close_dialog()
-                log.error("got unexpected stanza: "+str(stanza))
-                error = stanza.getTag("error")
-                if error and error.getTag("text"):
-                    ErrorDialog(_('Could not request upload slot'), 
-                                _('Got unexpected response from server: %s') % str(error.getTagData("text")),
-                                transient_for=self.chat_control.parent_win.window)
-                else:
-                    ErrorDialog(_('Could not request upload slot'), 
-                                _('Got unexpected response from server (protocol mismatch??)'),
-                                transient_for=self.chat_control.parent_win.window)
+                return
+
+            slot = stanza.getTag("slot")
+            if slot:
+                put = slot.getTag("put")
+                get = slot.getTag("get")
+            else:
+                progress_window.close_dialog()
+                log.error("got unexpected stanza: " + str(stanza))
+                ErrorDialog(_('Could not request upload slot'),
+                            _('Got unexpected response from server (see log)'),
+                            transient_for=self.chat_control.parent_win.window)
                 return
 
             try:
@@ -357,12 +379,12 @@ class Base(object):
                     iv = os.urandom(16)
                     data = StreamFileWithProgress(path_to_file,
                                                   "rb",
-                                                  progress_window.update_progress,
+                                                  progress_window.update_progress, event,
                                                   self.encrypted_upload, key, iv)
                 else:
                     data = StreamFileWithProgress(path_to_file,
                                                   "rb",
-                                                  progress_window.update_progress)
+                                                  progress_window.update_progress, event)
             except:
                 progress_window.close_dialog()
                 ErrorDialog(_('Could not open file'), 
@@ -370,19 +392,13 @@ class Base(object):
                             transient_for=self.chat_control.parent_win.window)
                 raise       # fill error log with useful information
 
-            put = slot.getTag("put")
-            get = slot.getTag("get")
-            if not put or not get:
-                progress_window.close_dialog()
-                log.error("got unexpected stanza: " + str(stanza))
-                ErrorDialog(_('Could not request upload slot'), 
-                            _('Got unexpected response from server (protocol mismatch??)'),
-                            transient_for=self.chat_control.parent_win.window)
-                return
-
             def upload_complete(response_code):
-                if response_code == 0:
-                    return      # Upload was aborted
+                if isinstance(response_code, str):
+                    # We got a error Message
+                    ErrorDialog(
+                        _('Error'), response_code,
+                        transient_for=self.chat_control.parent_win.window)
+                    return
                 if response_code >= 200 and response_code < 300:
                     log.info("Upload completed successfully")
                     xhtml = None
@@ -452,12 +468,12 @@ class Base(object):
                                     (get.getData(), get.getData(), thumb)
                     progress_window.close_dialog()
                     id_ = gajim.get_an_id()
-                    def add_oob_tag():
-                        pass
+
                     if self.encrypted_upload:
                         keyAndIv = '#' + binascii.hexlify(iv) + binascii.hexlify(key)
                         self.chat_control.send_message(message=get.getData() + keyAndIv, xhtml=None)
                     else:
+                        self.plugin.messages.append(get.getData())
                         self.chat_control.send_message(message=get.getData(), xhtml=xhtml)
                     self.chat_control.msg_textview.grab_focus()
                 else:
@@ -478,15 +494,17 @@ class Base(object):
                     transfer = urllib2.urlopen(request, timeout=30)
                     log.debug("urllib2 upload request done, response code: " + str(transfer.getcode()))
                     return transfer.getcode()
-                except UploadAbortedException:
-                    log.info("Upload aborted")
-                except:
-                    progress_window.close_dialog()
-                    ErrorDialog(_('Could not upload file'),
-                                _('Got unexpected exception while uploading file (see error log for more information)'),
-                                transient_for=self.chat_control.parent_win.window)
-                    raise       # fill error log with useful information
-                return 0
+                except UploadAbortedException as error:
+                    log.info('Upload Aborted')
+                    return str(error)
+                except urllib2.URLError as error:
+                    log.exception('URLError')
+                    errormsg = error.reason
+                except Exception as error:
+                    log.exception('Error')
+                    errormsg = error
+                gobject.idle_add(progress_window.close_dialog)
+                return str(errormsg)
 
             log.info("Uploading file to '%s'..." % str(put.getData()))
             log.info("Please download from '%s' later..." % str(get.getData()))
@@ -568,9 +586,10 @@ class Base(object):
 
 
 class StreamFileWithProgress(file):
-    def __init__(self, path, mode, callback=None,
+    def __init__(self, path, mode, callback, event,
                  encrypted_upload=False, key=None, iv=None, *args):
         file.__init__(self, path, mode)
+        self.event = event
         self.encrypted_upload = encrypted_upload
         self.seek(0, os.SEEK_END)
         if self.encrypted_upload:
@@ -594,6 +613,8 @@ class StreamFileWithProgress(file):
         return self._total
 
     def read(self, size):
+        if self.event.isSet():
+            raise UploadAbortedException
         if self.encrypted_upload:
             data = file.read(self, size)
             if len(data) > 0:
@@ -604,38 +625,33 @@ class StreamFileWithProgress(file):
                     data += self.encryptor.tag
                     self._seen += TAGSIZE
                 if self._callback:
-                    self._callback(self._seen, self._total, *self._args)
-            return data
+                    gobject.idle_add(self._callback, self._seen, self._total, *self._args)
         else:
             data = file.read(self, size)
             self._seen += len(data)
             if self._callback:
-                self._callback(self._seen, self._total, *self._args)
-            return data
+                gobject.idle_add(self._callback, self._seen, self._total, *self._args)
+        return data
 
 
 class ProgressWindow:
-    def __init__(self, title_text, during_text, messages_queue, plugin):
-        self.plugin = plugin
+    def __init__(self, during_text, messages_queue, base, event):
+        self.plugin = base.plugin
+        self.event = event
         self.xml = gtkgui_helpers.get_gtk_builder(self.plugin.local_file_path('upload_progress_dialog.ui'))
         self.messages_queue = messages_queue
         self.dialog = self.xml.get_object('progress_dialog')
+        self.dialog.set_transient_for(base.chat_control.parent_win.window)
         self.label = self.xml.get_object('label')
-        self.cancel_button = self.xml.get_object('close_button')
         self.label.set_markup('<big>' + during_text + '</big>')
         self.progressbar = self.xml.get_object('progressbar')
         self.progressbar.set_text("")
-        self.dialog.set_title(title_text)
-        self.dialog.set_geometry_hints(min_width=400, min_height=96)
-        self.dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
         self.dialog.show_all()
         self.xml.connect_signals(self)
 
-        self.stopped = False
         self.pulse_progressbar_timeout_id = gobject.timeout_add(100, self.pulse_progressbar)
         self.process_messages_queue_timeout_id = gobject.timeout_add(100, self.process_messages_queue)
 
-
     def pulse_progressbar(self):
         if self.dialog:
             self.progressbar.pulse()
@@ -649,22 +665,9 @@ class ProgressWindow:
             return True # loop forever
         return False
 
-    def on_progress_dialog_delete_event(self, widget, event):
-        self.stopped = True
-        if self.pulse_progressbar_timeout_id:
-            gobject.source_remove(self.pulse_progressbar_timeout_id)
-        gobject.source_remove(self.process_messages_queue_timeout_id)
-
-    def on_cancel(self, widget):
-        self.stopped = True
-        if self.pulse_progressbar_timeout_id:
-            gobject.source_remove(self.pulse_progressbar_timeout_id)
-        gobject.source_remove(self.process_messages_queue_timeout_id)
-        self.dialog.destroy()
-
     def update_progress(self, seen, total):
-        if self.stopped == True:
-            raise UploadAbortedException
+        if self.event.isSet():
+            return
         if self.pulse_progressbar_timeout_id:
             gobject.source_remove(self.pulse_progressbar_timeout_id)
             self.pulse_progressbar_timeout_id = None
@@ -673,8 +676,15 @@ class ProgressWindow:
         self.progressbar.set_text(str(int(pct)) + "%")
         log.debug('upload progress: %.2f%% (%d of %d bytes)' % (pct, seen, total))
 
-    def close_dialog(self):
-        self.on_cancel(None)
+    def close_dialog(self, *args):
+        self.dialog.destroy()
+
+    def on_destroy(self, event, *args):
+        self.event.set()
+        if self.pulse_progressbar_timeout_id:
+            gobject.source_remove(self.pulse_progressbar_timeout_id)
+        gobject.source_remove(self.process_messages_queue_timeout_id)
+
 
 class UploadAbortedException(Exception):
     def __str__(self):
